@@ -116,6 +116,20 @@ class OffsetScale(nn.Module):
         out = einsum('... d, h d -> ... h d', x, self.gamma) + self.beta
         return out.unbind(dim = -2)
 
+# activation functions
+
+class ReLUSquared(nn.Module):
+    def forward(self, x):
+        return F.relu(x) ** 2
+
+class LaplacianAttnFn(nn.Module):
+    """ https://arxiv.org/abs/2209.10655 claims this is more stable than Relu squared """
+
+    def forward(self, x):
+        mu = math.sqrt(0.5)
+        std = math.sqrt(0.25 * math.pi)
+        return (1 + torch.special.erf((x - mu) / (std * math.sqrt(2)))) * 0.5
+
 # gated attention unit
 
 class GAU(nn.Module):
@@ -128,6 +142,7 @@ class GAU(nn.Module):
         add_residual = True,
         causal = False,
         dropout = 0.,
+        laplace_attn_fn = False,
         norm_klass = nn.LayerNorm
     ):
         super().__init__()
@@ -137,6 +152,8 @@ class GAU(nn.Module):
         self.causal = causal
         self.dropout = nn.Dropout(dropout)
 
+        self.attn_fn = ReLUSquared() if not laplace_attn_fn else LaplacianAttnFn()
+
         self.to_hidden = nn.Sequential(
             nn.Linear(dim, hidden_dim * 2),
             nn.SiLU()
@@ -175,7 +192,7 @@ class GAU(nn.Module):
         if exists(rel_pos_bias):
             sim = sim + rel_pos_bias
 
-        attn = F.relu(sim) ** 2
+        attn = self.attn_fn(sim)
         attn = self.dropout(attn)
 
         if exists(mask):
@@ -210,7 +227,9 @@ class FLASH(nn.Module):
         dropout = 0.,
         rotary_pos_emb = None,
         norm_klass = nn.LayerNorm,
-        shift_tokens = False
+        shift_tokens = False,
+        laplace_attn_fn = False,
+        reduce_group_non_causal_attn = True
     ):
         super().__init__()
         hidden_dim = int(dim * expansion_factor)
@@ -218,6 +237,8 @@ class FLASH(nn.Module):
         self.causal = causal
         self.shift_tokens = shift_tokens
 
+        self.attn_fn = ReLUSquared() if not laplace_attn_fn else LaplacianAttnFn()
+
         # positional embeddings
 
         self.rotary_pos_emb = rotary_pos_emb
@@ -228,6 +249,10 @@ class FLASH(nn.Module):
         self.norm = norm_klass(dim)
         self.dropout = nn.Dropout(dropout)
 
+        # whether to reduce groups in non causal linear attention
+
+        self.reduce_group_non_causal_attn = reduce_group_non_causal_attn
+
         # projections
 
         self.to_hidden = nn.Sequential(
@@ -315,7 +340,7 @@ class FLASH(nn.Module):
 
         sim = sim + self.rel_pos_bias(sim)
 
-        attn = F.relu(sim) ** 2
+        attn = self.attn_fn(sim)
         attn = self.dropout(attn)
 
         if exists(mask):
@@ -339,8 +364,9 @@ class FLASH(nn.Module):
 
             lin_out = einsum('b g d e, b g n d -> b g n e', lin_kv, lin_q)
         else:
-            lin_kv = einsum('b g n d, b g n e -> b g d e', lin_k, v) / n
-            lin_out = einsum('b g n d, b g d e -> b g n e', lin_q, lin_kv)
+            context_einsum_eq = 'b d e' if self.reduce_group_non_causal_attn else 'b g d e'
+            lin_kv = einsum(f'b g n d, b g n e -> {context_einsum_eq}', lin_k, v) / n
+            lin_out = einsum(f'b g n d, {context_einsum_eq} -> b g n e', lin_q, lin_kv)
 
         # fold back groups into full sequence, and excise out padding
 
@@ -369,7 +395,9 @@ class FLASHTransformer(nn.Module):
         causal = False,
         attn_dropout = 0.,
         norm_type = 'scalenorm',
-        shift_tokens = True
+        shift_tokens = True,
+        laplace_attn_fn = False,
+        reduce_group_non_causal_attn = True
     ):
         super().__init__()
         assert norm_type in ('scalenorm', 'layernorm'), 'norm_type must be one of scalenorm or layernorm'
@@ -386,7 +414,7 @@ class FLASHTransformer(nn.Module):
         rotary_pos_emb = RotaryEmbedding(dim = min(32, query_key_dim))
         # max rotary embedding dimensions of 32, partial Rotary embeddings, from Wang et al - GPT-J
 
-        self.layers = nn.ModuleList([FLASH(dim = dim, group_size = group_size, query_key_dim = query_key_dim, expansion_factor = expansion_factor, causal = causal, dropout = attn_dropout, rotary_pos_emb = rotary_pos_emb, norm_klass = norm_klass, shift_tokens = shift_tokens) for _ in range(depth)])
+        self.layers = nn.ModuleList([FLASH(dim = dim, group_size = group_size, query_key_dim = query_key_dim, expansion_factor = expansion_factor, causal = causal, dropout = attn_dropout, rotary_pos_emb = rotary_pos_emb, norm_klass = norm_klass, shift_tokens = shift_tokens, reduce_group_non_causal_attn = reduce_group_non_causal_attn, laplace_attn_fn = laplace_attn_fn) for _ in range(depth)])
 
         self.to_logits = nn.Sequential(
             nn.LayerNorm(dim),
